#!/usr/bin/perl -T

use strict;
use warnings;

#*******************************************************************************
# Data::Filter::Factory class tests
#
# Author:   Jan Mach, jan.mach@cesnet.cz
# Version:  0.1
# Purpose:
# Comments:
#*******************************************************************************

use Data::Dumper;
#use Smart::Comments;

use Test::More tests => 232;

# First test if the module correctly compiles
BEGIN {
        use vars qw($TESTED_CLASS);
        $TESTED_CLASS = 'Data::Filter::Factory';

        use_ok('Tester');
        use_ok('Data::Filter::Rule::Constant');
        use_ok('Data::Filter::Rule::List');
        use_ok('Data::Filter::Rule::Variable');

        use_ok('Data::Filter::Rule::Existence');

        use_ok('Data::Filter::Rule::OperationUnary');
        use_ok('Data::Filter::Rule::OperationBinary');
        use_ok($TESTED_CLASS);
    }

# Debug output explicit ON switch
#Tester::DEBUG_ON();
#Tester::DEBUG_ON(10);

# Debug output explicit OFF switch
#Tester::DEBUG_OFF();

# User should know, which version is beeing tested (just for sure)
diag("$TESTED_CLASS version: " . $Data::Filter::Factory::VERSION) if Tester::D();

# Now attempt to create a class instance and check it
my $instance = $TESTED_CLASS->new_from_file('../../conf/filter-grammar.conf');
isa_ok($instance, $TESTED_CLASS);

# View the created instance
diag("Instance dump 01: " . Dumper($instance)) if Tester::D();

#-------------------------------------------------------------------------------
# Attempt to create few filter rules by hand, verify, that the factory method
# can create all rule classess
#-------------------------------------------------------------------------------

my $f1 = Data::Filter::Factory->rule('const', 'const_value');
isa_ok($f1, 'Data::Filter::Rule');
is($f1->to_string(), '"const_value"', 'Filter test 01');
diag("F01: " . Dumper($f1) . "\n") if Tester::D();

my $f2 = Data::Filter::Factory->rule('list', '"value1","value2"');
isa_ok($f2, 'Data::Filter::Rule');
is($f2->to_string(), '["value1","value2"]', 'Filter test 02');
diag("F02: " . Dumper($f2) . "\n") if Tester::D();

my $f3 = Data::Filter::Factory->rule('list', ["value1","value2"]);
isa_ok($f3, 'Data::Filter::Rule');
is($f3->to_string(), '["value1","value2"]', 'Filter test 03');
diag("F03: " . Dumper($f3) . "\n") if Tester::D();

my $f4 = Data::Filter::Factory->rule('var', 'Alert/@messageid');
isa_ok($f4, 'Data::Filter::Rule') if Tester::D();
is($f4->to_string(), 'Alert/@messageid', 'Filter test 04');
diag("F04: " . Dumper($f4) . "\n") if Tester::D();

my $f5 = Data::Filter::Factory->rule('exists', 'Alert/@messageid');
isa_ok($f5, 'Data::Filter::Rule');
is($f5->to_string(), 'EXISTS Alert/@messageid', 'Manual filter test 05');
diag("F05: " . Dumper($f5) . "\n") if Tester::D();

my $f6 = Data::Filter::Factory->rule('eq', $f4, $f1);
isa_ok($f6, 'Data::Filter::Rule');
is($f6->to_string(), '(Alert/@messageid EQ "const_value")', 'Manual filter test 06');
diag("F06: " . Dumper($f6) . "\n") if Tester::D();

my $f7 = Data::Filter::Factory->rule('not', $f5);
isa_ok($f7, 'Data::Filter::Rule');
is($f7->to_string(), '(NOT EXISTS Alert/@messageid)', 'Manual filter test 07');
diag("F7: " . Dumper($f7) . "\n") if Tester::D();

#-------------------------------------------------------------------------------

my ($counter, $filter);

#-------------------------------------------------------------------------------
# Test automatic constant compilation
#-------------------------------------------------------------------------------

my @RULES_C = (
        [
            'const', [123456, 'NUMERIC'], 'Data::Filter::Rule::Constant',
            '"123456"',
        ],
        [
            'const', [123.456, 'NUMERIC'], 'Data::Filter::Rule::Constant',
            '"123.456"',
        ],
        [
            'const', [0.1234, 'NUMERIC'], 'Data::Filter::Rule::Constant',
            '"0.1234"',
        ],
        [
            'const', ['0xabcd', 'NUMERIC'], 'Data::Filter::Rule::Constant',
            '"43981"',
        ],
        [
            'const', ['012345', 'NUMERIC'], 'Data::Filter::Rule::Constant',
            '"5349"',
        ],
        [
            'const', ['0b1111', 'NUMERIC'], 'Data::Filter::Rule::Constant',
            '"15"',
        ],
        #-----------------------------------------------------------------------
        [
            'const', ['127.1.1.1', 'ADDRESS'], 'Data::Filter::Rule::Constant',
            '"127.1.1.1"',
        ],
        #-----------------------------------------------------------------------
        [
            'list', [['127.1.1.1','127.1.1.2'], 'ADDRESS'], 'Data::Filter::Rule::List',
            '["127.1.1.1","127.1.1.2"]',
        ],
        #-----------------------------------------------------------------------
    );

$counter = 0;
foreach my $rule (@RULES_C) {
    ++$counter;
    $filter = Data::Filter::Factory->rule($rule->[0], @{$rule->[1]});
    isa_ok($filter, $rule->[2]);
    is($filter->to_string(), $rule->[3], "Filter test A$counter");
    diag("F$counter: " . Dumper($filter) . "\n") if Tester::D();
}

#-------------------------------------------------------------------------------
# Create filters by using rules
#-------------------------------------------------------------------------------

# FILTER RULES TO BE TESTED IN STRUCTURE WITH FOLLOWING SYNTAX
#  index 0 - filter rule
#  index 1 - class of the expected top rule tree element
#  index 2 - expected filter structure (as printed by to_string() method)
# NOTE:
#  If the rule should not be parsed (testing invalid rule to be really invalid),
#  set the index 1 and 2 values to 'undef'.

my @RULES_F = (
        #-----------------------------------------------------------------------
        # Valid filter rules (valid syntax, must parse)
        #-----------------------------------------------------------------------

        # Check, that all DATETIME constants are correctly recognized
        ['2013-02-04T13:52',         'Data::Filter::Rule::Constant', '"2013-02-04 12:52:00Z"'],
        ['2013-02-04T13:52:56',      'Data::Filter::Rule::Constant', '"2013-02-04 12:52:56Z"'],
        ['2013-02-04T13:52Z',        'Data::Filter::Rule::Constant', '"2013-02-04 13:52:00Z"'],
        ['2013-02-04T13:52:56Z',     'Data::Filter::Rule::Constant', '"2013-02-04 13:52:56Z"'],
        ['2013-02-04T13:52+0100',    'Data::Filter::Rule::Constant', '"2013-02-04 12:52:00Z"'],
        ['2013-02-04T13:52:56+0100', 'Data::Filter::Rule::Constant', '"2013-02-04 12:52:56Z"'],
        ['2013-02-04T13:52-0100',    'Data::Filter::Rule::Constant', '"2013-02-04 14:52:00Z"'],
        ['2013-02-04T13:52:56-0100', 'Data::Filter::Rule::Constant', '"2013-02-04 14:52:56Z"'],

        # Check, that all IPV4 constants are correctly recognized
        ['127.0.0.1',            'Data::Filter::Rule::Constant', '"127.0.0.1"'],
        ['127.0.0.1/32',         'Data::Filter::Rule::Constant', '"127.0.0.1/32"'],
        ['127.0.0.1-127.0.0.2',  'Data::Filter::Rule::Constant', '"127.0.0.1..127.0.0.2"'],
        ['127.0.0.1..127.0.0.2', 'Data::Filter::Rule::Constant', '"127.0.0.1..127.0.0.2"'],

        # Check, that all IPV6 constants are correctly recognized
        ['::',                         'Data::Filter::Rule::Constant', '"::"'],
        ['::1',                        'Data::Filter::Rule::Constant', '"::1"'],
        ['2012:abcd::1',               'Data::Filter::Rule::Constant', '"2012:ABCD::1"'],
        #['::/64',                      'Data::Filter::Rule::Constant', '"::/64"'],
        #['::1/64',                     'Data::Filter::Rule::Constant', '"::1/64"'],
        #['2012:abcd::1/64',            'Data::Filter::Rule::Constant', '"2012:abcd::1/64"'],
        #['::-::5',                     'Data::Filter::Rule::Constant', '"::-::5"'],
        #['::1-::5',                    'Data::Filter::Rule::Constant', '"::1-::5"'],
        #['2012:abcd::1-2012:abcd::5',  'Data::Filter::Rule::Constant', '"2012:abcd::1-2012:abcd::5"'],
        #['::..::5',                    'Data::Filter::Rule::Constant', '"::..::5"'],
        #['::1..::5',                   'Data::Filter::Rule::Constant', '"::1..::5"'],
        #['2012:abcd::1..2012:abcd::5', 'Data::Filter::Rule::Constant', '"2012:abcd::1..2012:abcd::5"'],

        # Check, that all NUMERIC constants are correctly recognized
        ['123456789', 'Data::Filter::Rule::Constant', '"123456789"'],
        ['1234.5678', 'Data::Filter::Rule::Constant', '"1234.5678"'],
        ['0.1234567', 'Data::Filter::Rule::Constant', '"0.1234567"'],
        ['0x12abAB',  'Data::Filter::Rule::Constant', '"1223595"'],
        ['01234567',  'Data::Filter::Rule::Constant', '"342391"'],
        ['0b11111',   'Data::Filter::Rule::Constant', '"31"'],

        # Check, that all lists are correctly recognized
        ['[item1,item2]',   'Data::Filter::Rule::List', '["item1","item2"]'],
        ['["item1",item2]', 'Data::Filter::Rule::List', '["item1","item2"]'],
        ['["item1",{}()]',  'Data::Filter::Rule::List', '["item1","{}()"]'],

        # Check, that all quoted constants are correctly recognized
        ['"constant"',             'Data::Filter::Rule::Constant', '"constant"'],
        ['"constant with spaces"', 'Data::Filter::Rule::Constant', '"constant with spaces"'],
        ['"constant with []{}()"', 'Data::Filter::Rule::Constant', '"constant with []{}()"'],

        # Check, that all MPATH definitions are correctly recognized
        ['Alert',            'Data::Filter::Rule::Variable', 'Alert'],
        ['Alert/@messageid', 'Data::Filter::Rule::Variable', 'Alert/@messageid'],

        # Check all variants of comparison operations
        ['Node/Name == "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name EQ "a1b2c3d5")'],
        ['Node/Name eq "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name EQ "a1b2c3d5")'],
        ['Node/Name EQ "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name EQ "a1b2c3d5")'],
        ['Node/Name != "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name NE "a1b2c3d5")'],
        ['Node/Name ne "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name NE "a1b2c3d5")'],
        ['Node/Name NE "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name NE "a1b2c3d5")'],
        ['Node/Name <  "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name LT "a1b2c3d5")'],
        ['Node/Name lt "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name LT "a1b2c3d5")'],
        ['Node/Name LT "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name LT "a1b2c3d5")'],
        ['Node/Name >  "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name GT "a1b2c3d5")'],
        ['Node/Name gt "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name GT "a1b2c3d5")'],
        ['Node/Name GT "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name GT "a1b2c3d5")'],
        ['Node/Name <= "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name LE "a1b2c3d5")'],
        ['Node/Name le "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name LE "a1b2c3d5")'],
        ['Node/Name LE "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name LE "a1b2c3d5")'],
        ['Node/Name >= "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name GE "a1b2c3d5")'],
        ['Node/Name ge "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name GE "a1b2c3d5")'],
        ['Node/Name GE "a1b2c3d5"',  'Data::Filter::Rule::OperationBinary', '(Node/Name GE "a1b2c3d5")'],
        ['Node/Name =~ "^a\db\d"',   'Data::Filter::Rule::OperationBinary', '(Node/Name LIKE "^a\db\d")'],
        ['Node/Name like "^a\db\d"', 'Data::Filter::Rule::OperationBinary', '(Node/Name LIKE "^a\db\d")'],
        ['Node/Name LIKE "^a\db\d"', 'Data::Filter::Rule::OperationBinary', '(Node/Name LIKE "^a\db\d")'],
        ['Node/Name ~~ ["a1","b2"]', 'Data::Filter::Rule::OperationBinary', '(Node/Name IN ["a1","b2"])'],
        ['Node/Name in ["a1","b2"]', 'Data::Filter::Rule::OperationBinary', '(Node/Name IN ["a1","b2"])'],
        ['Node/Name IN ["a1","b2"]', 'Data::Filter::Rule::OperationBinary', '(Node/Name IN ["a1","b2"])'],

        # Check existence operator
        ['? Node/Name',          'Data::Filter::Rule::Existence', 'EXISTS Node/Name'],
        ['exists Node/Name',     'Data::Filter::Rule::Existence', 'EXISTS Node/Name'],
        ['EXISTS Node/Name',     'Data::Filter::Rule::Existence', 'EXISTS Node/Name'],

        # Check all variants of logical operations and also check, that comparison
        # operations take precedence from logical operations
        [
            'Node/Name == "a1b2c3d5" && Source/IP4 eq 192.168.0.1',
            'Data::Filter::Rule::OperationBinary',
            '((Node/Name EQ "a1b2c3d5") AND (Source/IP4 EQ "192.168.0.1"))'
        ],
        [
            'Node/Name == "a1b2c3d5" and Source/IP4 eq 192.168.0.1',
            'Data::Filter::Rule::OperationBinary',
            '((Node/Name EQ "a1b2c3d5") AND (Source/IP4 EQ "192.168.0.1"))'
        ],
        [
            'Node/Name == "a1b2c3d5" AND Source/IP4 eq 192.168.0.1',
            'Data::Filter::Rule::OperationBinary',
            '((Node/Name EQ "a1b2c3d5") AND (Source/IP4 EQ "192.168.0.1"))'
        ],
        [
            'Node/Name == "a1b2c3d5" || Source/IP4 eq 192.168.0.1',
            'Data::Filter::Rule::OperationBinary',
            '((Node/Name EQ "a1b2c3d5") OR (Source/IP4 EQ "192.168.0.1"))'
        ],
        [
            'Node/Name == "a1b2c3d5" or Source/IP4 eq 192.168.0.1',
            'Data::Filter::Rule::OperationBinary',
            '((Node/Name EQ "a1b2c3d5") OR (Source/IP4 EQ "192.168.0.1"))'
        ],
        [
            'Node/Name == "a1b2c3d5" OR Source/IP4 eq 192.168.0.1',
            'Data::Filter::Rule::OperationBinary',
            '((Node/Name EQ "a1b2c3d5") OR (Source/IP4 EQ "192.168.0.1"))'
        ],
        [
            'Node/Name == "a1b2c3d5" ^^ Source/IP4 eq 192.168.0.1',
            'Data::Filter::Rule::OperationBinary',
            '((Node/Name EQ "a1b2c3d5") XOR (Source/IP4 EQ "192.168.0.1"))'
        ],
        [
            'Node/Name == "a1b2c3d5" xor Source/IP4 eq 192.168.0.1',
            'Data::Filter::Rule::OperationBinary',
            '((Node/Name EQ "a1b2c3d5") XOR (Source/IP4 EQ "192.168.0.1"))'
        ],
        [
            'Node/Name == "a1b2c3d5" XOR Source/IP4 eq 192.168.0.1',
            'Data::Filter::Rule::OperationBinary',
            '((Node/Name EQ "a1b2c3d5") XOR (Source/IP4 EQ "192.168.0.1"))'
        ],

        # Check, that existence precedes all logical operations
        [
            'EXISTS Node/Name AND Source/IP4 eq 192.168.0.1',
            'Data::Filter::Rule::OperationBinary',
            '(EXISTS Node/Name AND (Source/IP4 EQ "192.168.0.1"))'
        ],
        [
            'EXISTS Node/Name OR Source/IP4 eq 192.168.0.1',
            'Data::Filter::Rule::OperationBinary',
            '(EXISTS Node/Name OR (Source/IP4 EQ "192.168.0.1"))'
        ],
        [
            'EXISTS Node/Name XOR Source/IP4 eq 192.168.0.1',
            'Data::Filter::Rule::OperationBinary',
            '(EXISTS Node/Name XOR (Source/IP4 EQ "192.168.0.1"))'
        ],

        # Check, that XOR precedes OR (left associative)
        [
            'Node/Name == "a1b2c3d5" OR Source/IP4 eq 192.168.0.1 XOR CreateTime ge 2013-04-18T10:28:37Z',
            'Data::Filter::Rule::OperationBinary',
            '((Node/Name EQ "a1b2c3d5") OR ((Source/IP4 EQ "192.168.0.1") XOR (CreateTime GE "2013-04-18 10:28:37Z")))'
        ],
        # Check, that AND precedes OR (left associative)
        [
            'Node/Name == "a1b2c3d5" OR Source/IP4 eq 192.168.0.1 AND CreateTime ge 2013-04-18T10:28:37Z',
            'Data::Filter::Rule::OperationBinary',
            '((Node/Name EQ "a1b2c3d5") OR ((Source/IP4 EQ "192.168.0.1") AND (CreateTime GE "2013-04-18 10:28:37Z")))'
        ],
        # Check, that AND precedes XOR (left associative)
        [
            'Node/Name == "a1b2c3d5" XOR Source/IP4 eq 192.168.0.1 AND CreateTime ge 2013-04-18T10:28:37Z',
            'Data::Filter::Rule::OperationBinary',
            '((Node/Name EQ "a1b2c3d5") XOR ((Source/IP4 EQ "192.168.0.1") AND (CreateTime GE "2013-04-18 10:28:37Z")))'
        ],
        # Check, that XOR precedes OR (right associative)
        [
            'Node/Name == "a1b2c3d5" XOR Source/IP4 eq 192.168.0.1 OR CreateTime ge 2013-04-18T10:28:37Z',
            'Data::Filter::Rule::OperationBinary',
            '(((Node/Name EQ "a1b2c3d5") XOR (Source/IP4 EQ "192.168.0.1")) OR (CreateTime GE "2013-04-18 10:28:37Z"))'
        ],
        # Check, that AND precedes OR (right associative)
        [
            'Node/Name == "a1b2c3d5" AND Source/IP4 eq 192.168.0.1 OR CreateTime ge 2013-04-18T10:28:37Z',
            'Data::Filter::Rule::OperationBinary',
            '(((Node/Name EQ "a1b2c3d5") AND (Source/IP4 EQ "192.168.0.1")) OR (CreateTime GE "2013-04-18 10:28:37Z"))'
        ],
        # Check, that AND precedes XOR (right associative)
        [
            'Node/Name == "a1b2c3d5" AND Source/IP4 eq 192.168.0.1 XOR CreateTime ge 2013-04-18T10:28:37Z',
            'Data::Filter::Rule::OperationBinary',
            '(((Node/Name EQ "a1b2c3d5") AND (Source/IP4 EQ "192.168.0.1")) XOR (CreateTime GE "2013-04-18 10:28:37Z"))'
        ],

        # Check, that NOT beats everything
        [
            'NOT Node/Name == "a1b2c3d5" OR NOT Source/IP4 eq 192.168.0.1 XOR NOT CreateTime ge 2013-04-18T10:28:37Z',
            'Data::Filter::Rule::OperationBinary',
            '((NOT (Node/Name EQ "a1b2c3d5")) OR ((NOT (Source/IP4 EQ "192.168.0.1")) XOR (NOT (CreateTime GE "2013-04-18 10:28:37Z"))))'
        ],
        # Check, that AND precedes OR (left associative)
        [
            'NOT Node/Name == "a1b2c3d5" OR NOT Source/IP4 eq 192.168.0.1 AND NOT CreateTime ge 2013-04-18T10:28:37Z',
            'Data::Filter::Rule::OperationBinary',
            '((NOT (Node/Name EQ "a1b2c3d5")) OR ((NOT (Source/IP4 EQ "192.168.0.1")) AND (NOT (CreateTime GE "2013-04-18 10:28:37Z"))))'
        ],
        # Check, that AND precedes XOR (left associative)
        [
            'NOT Node/Name == "a1b2c3d5" XOR NOT Source/IP4 eq 192.168.0.1 AND NOT CreateTime ge 2013-04-18T10:28:37Z',
            'Data::Filter::Rule::OperationBinary',
            '((NOT (Node/Name EQ "a1b2c3d5")) XOR ((NOT (Source/IP4 EQ "192.168.0.1")) AND (NOT (CreateTime GE "2013-04-18 10:28:37Z"))))'
        ],
        # Check, that XOR precedes OR (right associative)
        [
            'NOT Node/Name == "a1b2c3d5" XOR NOT Source/IP4 eq 192.168.0.1 OR NOT CreateTime ge 2013-04-18T10:28:37Z',
            'Data::Filter::Rule::OperationBinary',
            '(((NOT (Node/Name EQ "a1b2c3d5")) XOR (NOT (Source/IP4 EQ "192.168.0.1"))) OR (NOT (CreateTime GE "2013-04-18 10:28:37Z")))'
        ],
        # Check, that AND precedes OR (right associative)
        [
            'NOT Node/Name == "a1b2c3d5" AND NOT Source/IP4 eq 192.168.0.1 OR NOT CreateTime ge 2013-04-18T10:28:37Z',
            'Data::Filter::Rule::OperationBinary',
            '(((NOT (Node/Name EQ "a1b2c3d5")) AND (NOT (Source/IP4 EQ "192.168.0.1"))) OR (NOT (CreateTime GE "2013-04-18 10:28:37Z")))'
        ],
        # Check, that AND precedes XOR (right associative)
        [
            'NOT Node/Name == "a1b2c3d5" AND NOT Source/IP4 eq 192.168.0.1 XOR NOT CreateTime ge 2013-04-18T10:28:37Z',
            'Data::Filter::Rule::OperationBinary',
            '(((NOT (Node/Name EQ "a1b2c3d5")) AND (NOT (Source/IP4 EQ "192.168.0.1"))) XOR (NOT (CreateTime GE "2013-04-18 10:28:37Z")))'
        ],

        # Check mathematical operators
        ['1 + 2', 'Data::Filter::Rule::OperationBinary', '("1" ADD "2")'],
        ['1 - 2', 'Data::Filter::Rule::OperationBinary', '("1" SUB "2")'],
        ['1 * 2', 'Data::Filter::Rule::OperationBinary', '("1" MUL "2")'],
        ['1 / 2', 'Data::Filter::Rule::OperationBinary', '("1" DIV "2")'],
        ['1 % 2', 'Data::Filter::Rule::OperationBinary', '("1" MOD "2")'],
        ['1 . 2', 'Data::Filter::Rule::OperationBinary', '("1" CON "2")'],

        # Check, that mathematical operators take precedence over anything
        [
            'NOT 1 + 2 AND NOT 2 + 1 eq 192.168.0.1 XOR NOT CreateTime ge 6 + 7',
            'Data::Filter::Rule::OperationBinary',
            '(((NOT ("1" ADD "2")) AND (NOT (("2" ADD "1") EQ "192.168.0.1"))) XOR (NOT (CreateTime GE ("6" ADD "7"))))'
        ],

        # Check logical operation precedence
        [
            'DetectTime GE "2016-09-26 02:00:00" and Node/Name IN ["a","b","c","d","e"]',
            'Data::Filter::Rule::OperationBinary',
            '((DetectTime GE "2016-09-26 02:00:00") AND (Node/Name IN ["a","b","c","d","e"]))'
        ],
        [
            'DetectTime GE "2016-09-26 02:00:00" AND Node/Name IN ["a","b","c","d","e"]',
            'Data::Filter::Rule::OperationBinary',
            '((DetectTime GE "2016-09-26 02:00:00") AND (Node/Name IN ["a","b","c","d","e"]))'
        ],
        [
            'DetectTime GE "2016-09-26 02:00:00" && Node/Name IN ["a","b","c","d","e"]',
            'Data::Filter::Rule::OperationBinary',
            '((DetectTime GE "2016-09-26 02:00:00") AND (Node/Name IN ["a","b","c","d","e"]))'
        ],
        [
            'DetectTime GE "2016-09-26 02:00:00" or Node/Name IN ["a","b","c","d","e"]',
            'Data::Filter::Rule::OperationBinary',
            '((DetectTime GE "2016-09-26 02:00:00") OR (Node/Name IN ["a","b","c","d","e"]))'
        ],
        [
            'DetectTime GE "2016-09-26 02:00:00" OR Node/Name IN ["a","b","c","d","e"]',
            'Data::Filter::Rule::OperationBinary',
            '((DetectTime GE "2016-09-26 02:00:00") OR (Node/Name IN ["a","b","c","d","e"]))'
        ],
        [
            'DetectTime GE "2016-09-26 02:00:00" || Node/Name IN ["a","b","c","d","e"]',
            'Data::Filter::Rule::OperationBinary',
            '((DetectTime GE "2016-09-26 02:00:00") OR (Node/Name IN ["a","b","c","d","e"]))'
        ],
        [
            'DetectTime GE "2016-09-26 02:00:00" AND Source/IP4/ip EQ "89.163.148.231" OR Target/IP4/ip EQ "147.230.18.73" AND Category IN ["Attempt.Exploit"]',
            'Data::Filter::Rule::OperationBinary',
            '(((DetectTime GE "2016-09-26 02:00:00") AND (Source/IP4/ip EQ "89.163.148.231")) OR ((Target/IP4/ip EQ "147.230.18.73") AND (Category IN ["Attempt.Exploit"])))'
        ],
        [
            'DetectTime GE "2016-09-26 02:00:00" AND Source/IP4/ip EQ "89.163.148.231" || Target/IP4/ip EQ "147.230.18.73" AND Category IN ["Attempt.Exploit"]',
            'Data::Filter::Rule::OperationBinary',
            '((DetectTime GE "2016-09-26 02:00:00") AND (((Source/IP4/ip EQ "89.163.148.231") OR (Target/IP4/ip EQ "147.230.18.73")) AND (Category IN ["Attempt.Exploit"])))'
        ],

        #-----------------------------------------------------------------------
        # Invalid filter rules (invalid syntax, must not parse)
        #-----------------------------------------------------------------------

        ['ExistS Alert',     undef],
        ['? 12345',          undef],
        ['exists "Alert"',   undef],
        ['EXISTS 127.0.0.1', undef],
    );

$counter = 0;
foreach my $rule (@RULES_F) {
    ++$counter;
    eval {
        $filter = undef;
        $filter = $instance->filter_from_string($rule->[0]);
    };
    if (defined $rule->[1]) {
        isa_ok($filter, $rule->[1]);
        if ($filter) {
            is($filter->to_string(), $rule->[2], "Valid filter test $counter: $rule->[0]");
            diag("FILTER $counter: " . Dumper($filter) . "\n") if Tester::D();
        }
        else {
            fail("Failed valid filter test $counter: $rule->[0] $@");
        }
    }
    else {
        is($filter, undef, "Invalid filter test $counter")
    }
}
