#!/usr/bin/perl -T

use strict;
use warnings;

#*******************************************************************************
# Data::Tree class tests
#
# Author:   Jan Mach, jan.mach@cesnet.cz
# Version:  0.1
# Purpose:
# Comments:
#*******************************************************************************

use Data::Dumper;
#use Smart::Comments;

use Test::More tests => 17;

# First test if the module correctly compiles
BEGIN {
        use vars qw($TESTED_CLASS);
        $TESTED_CLASS = 'Data::Tree';

        use_ok('Tester');
        use_ok('Data::Path::Parser');
        use_ok($TESTED_CLASS);
    }

# Debug output explicit ON switch
#Tester::DEBUG_ON();

# Debug output explicit OFF switch
#Tester::DEBUG_OFF();

# User should know, which version is beeing tested (just for sure)
diag("$TESTED_CLASS version: " . $Data::Tree::VERSION) if Tester::D();

# Now attempt to create a class instance and check it
my $instance = $TESTED_CLASS->new();
isa_ok($instance, $TESTED_CLASS);

# View the created instance
diag("Instance dump 01: " . Dumper($instance)) if Tester::D();

is($instance->to_string(1),
'{}
', "Instance print test 01");

# Check existence of some nodes
is($instance->path_node_exists('Node/Name'), 0, 'Existence test 01');
is($instance->path_node_exists('Category'),  0, 'Existence test 02');

my @MESSAGE_RULES = (
    ["ID", "2E4A3926-B1B9-41E3-89AE-B6B474EB0A54"],
    ["DetectTime", "2014-03-22T10:12:31Z"],
    ["Category[1]", "Recon.Scanning"],
    ["ConnCount", 633],
    ["Description", "EPMAPPER exploitation attempt"],
    ["Ref[1]", "cve:CVE-2003-0605"],
    ["Source[1]/IP4[1]", "93.184.216.119"],
    ["Source[1]/Proto[1]", "tcp"],
    ["Source[1]/Proto[2]", "epmap"],
    ["Source[1]/Port[1]", "24508"],
    ["Target[1]/Proto[1]", "tcp"],
    ["Target[1]/Proto[2]", "epmap"],
    ["Target[1]/Port[1]", "135"],
    ["Node[1]/Name", "kippo-honey"],
    ["Node[1]/Realm", "cesnet.cz"],
    ["Node[1]/Tags[1]", "Protocol"],
    ["Node[1]/Tags[2]", "Honeypot"],
    ["Node[1]/SW", "Kippo"],
    ["Node[1]/AggrWin", "00:05:00"],
);

# Create some message structure
foreach my $rule (@MESSAGE_RULES) {
    $instance->path_node_set(@$rule);
}

# Create some message structure
foreach my $rule (@MESSAGE_RULES) {
    $instance->path_node_set(@$rule);
}

my $expected = {
    "ID" => "2E4A3926-B1B9-41E3-89AE-B6B474EB0A54",
    "Description" => "EPMAPPER exploitation attempt",
    "Category" => ["Recon.Scanning"],
    "DetectTime" => "2014-03-22T10:12:31Z",
    "Ref" => ["cve:CVE-2003-0605"],
    "ConnCount" => 633,
    "Source" => [
        {
            "Proto" => [
                "tcp",
                "epmap"
            ],
            "Port" => ["24508"],
            "IP4" => ["93.184.216.119"],
        }
    ],
    "Target" => [
        {
            "Port" => ["135"],
            "Proto" => [
                "tcp",
                "epmap"
            ]
        }
    ],
    "Node" => [
        {
            "SW" => "Kippo",
            "AggrWin" => "00:05:00",
            "Realm" => "cesnet.cz",
            "Tags" => [
                "Protocol",
                "Honeypot"
            ],
        "Name" => "kippo-honey"
        }
    ],
};

# View the message structure
diag("Instance dump 02: " . Dumper($instance)) if Tester::D();
is_deeply($instance->hash(), $expected, "Instance dump test 01");


diag("'Source/IP4' result: " . Dumper($instance->path_nodes('Source/IP4'))) if Tester::D();
diag("'Target' result: " . Dumper($instance->path_nodes('Target')))  if Tester::D();
diag("'Node' result: " . Dumper($instance->path_values('Node'))) if Tester::D();

#my $nodes = $instance->path_nodes('Source/IP4');
#foreach my $n (@$nodes) {
#   $$n = 'xx.xx.xx.xx';
#}
#diag("Instance dump XX: " . Dumper($instance)) if Tester::D();

# Test simple iteration
my ($mpath, $value, $node);
($mpath, $value) = $instance->iterate();
do {
    diag("ITERATE A: $mpath => $value") if Tester::D();
    ($mpath, $value) = $instance->iterate();
} while $mpath;

# Test iteration in "NODE" mode (returns references, so you may change the values)
($mpath, $node) = $instance->iterate(undef,1);
do {
    diag("ITERATE B: $mpath => $node") if Tester::D();
    #$$node = 'test'; # Alter the current value to new one
    ($mpath, $node) = $instance->iterate(undef,1);
} while $mpath;

# Check the changes made in previous step
($mpath, $value) = $instance->iterate();
do {
    diag("ITERATE C: $mpath => $value") if Tester::D();
    ($mpath, $value) = $instance->iterate();
} while $mpath;

# Suply the MPath pattern to iterate only over certain nodes
($mpath, $value) = $instance->iterate('Proto');
do {
    diag("ITERATE D: $mpath => $value") if Tester::D();
    ($mpath, $value) = $instance->iterate('Proto');
} while $mpath;

# These calls should now yield 1
is($instance->path_node_exists('Node/Name'), 1, 'Existence test 03');
is($instance->path_node_exists('Category'),  1, 'Existence test 04');

# Retrieve some values
is($instance->path_value('Node/Name'), 'kippo-honey',    'Get node value test 01');
is($instance->path_value('Category'),  'Recon.Scanning', 'Get node value test 02');

# Retrieve multiple values
is_deeply($instance->path_values('Node/Tags'), ['Protocol', 'Honeypot'], 'Get multiple node values test 01');

# Change some values
$instance->path_value_set('Node/Name',   'kippo-honey-updated');
$instance->path_value_set('Category[1]', 'Recon.Scanning-updated');

# Verify the values are new
is($instance->path_value('Node/Name'), 'kippo-honey-updated',    'Get node value test 03');
is($instance->path_value('Category'),  'Recon.Scanning-updated', 'Get node value test 04');

# View the message structure
diag("Instance dump 03: " . Dumper($instance)) if Tester::D();

# Delete some nodes
$instance->path_node_delete('Node/Name');
$instance->path_node_delete('Category[1]');

# Verify the nodes are gone
is($instance->path_node_exists('Node/Name'), 0, 'Existence test 05');
is($instance->path_node_exists('Category'),  0, 'Existence test 06');

# View the message structure
diag("Instance dump 04: " . Dumper($instance)) if Tester::D();

# Test the message serialization and deserialization
my $serialization = $instance->serialize();
my $instance2     = Data::Tree->unserialize($serialization);

# View the message structure after deserialization
diag("Instance dump 05: " . Dumper($instance2)) if Tester::D();
